// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "Mesh.hpp"

using namespace Dawn;

void Mesh::createBuffers(
  int32_t verticeCount,
  int32_t indiceCount
) {
  if(verticeCount <= 0) throw "Vertice count must be greater than zero.";
  if(indiceCount <= 0) throw "Indice count must be greater than zero.";

  this->disposeBuffers();

  this->verticeCount = verticeCount;
  this->indiceCount = indiceCount;

  auto sizePos = sizeof(glm::vec3) * verticeCount;
  auto sizeInds = sizeof(meshindice_t) * indiceCount;
  auto sizeCoords = sizeof(glm::vec2) * verticeCount;

  // Create some buffers, one for the vertex data, one for the indices
  GLuint buffer[2];
  glGenBuffers(2, buffer);
  this->vertexBuffer = buffer[0];
  if(this->vertexBuffer < 0) throw "Failed to create vertex buffer";
  this->indexBuffer = buffer[1];
  if(this->indexBuffer < 0) throw "Failed to create index buffer";

  // Buffer an empty set of data then buffer each component
  glBindBuffer(GL_ARRAY_BUFFER, this->vertexBuffer);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, this->indexBuffer);
  glBufferData(GL_ARRAY_BUFFER, sizePos+sizeCoords, 0, GL_DYNAMIC_DRAW);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeInds, 0, GL_DYNAMIC_DRAW);

  // Setup the attrib pointers
  size_t offset = 0;
  glVertexAttribPointer(
    0, sizeof(glm::vec3) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  glEnableVertexAttribArray(0);

  offset += sizePos;
  glVertexAttribPointer(
    1, sizeof(glm::vec2) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  glEnableVertexAttribArray(1);
}

void Mesh::disposeBuffers() {
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

  if(this->vertexBuffer != -1) {
    glDeleteBuffers(1, &this->vertexBuffer);
    this->vertexBuffer = -1;
    this->verticeCount = -1;
  }

  if(this->indexBuffer != -1) {
    glDeleteBuffers(1, &this->indexBuffer);
    this->indexBuffer = -1;
    this->indiceCount = -1;
  }
}

void Mesh::bufferPositions(int32_t pos, glm::vec3 *positions, int32_t len) {
  assertNotNull(positions);
  glBindBuffer(GL_ARRAY_BUFFER, this->vertexBuffer);
  glBufferSubData(
    GL_ARRAY_BUFFER,
    sizeof(glm::vec3) * pos,
    sizeof(glm::vec3) * len,
    (void*)positions
  );
}

void Mesh::bufferCoordinates(int32_t pos, glm::vec2 *coordinates, int32_t len) {
  assertNotNull(coordinates);

  auto offsetCoordinates = (
    (sizeof(glm::vec3) * this->verticeCount) +
    (sizeof(glm::vec2) * pos)
  );

  glBindBuffer(GL_ARRAY_BUFFER, this->vertexBuffer);
  glBufferSubData(
    GL_ARRAY_BUFFER,
    offsetCoordinates,
    sizeof(glm::vec2) * len,
    (void*)coordinates
  );
}

void Mesh::bufferIndices(int32_t pos, meshindice_t *indices, int32_t len) {
  assertNotNull(indices);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, this->indexBuffer);
  glBufferSubData(
    GL_ELEMENT_ARRAY_BUFFER,
    sizeof(meshindice_t) * pos,
    sizeof(meshindice_t) * len,
    (void*)indices
  );
}

void Mesh::draw(enum MeshDrawMode drawMode, int32_t start, int32_t count) {
  if(
    count == 0 ||
    this->vertexBuffer == -1 ||
    this->indexBuffer == -1
  ) return;

  if(count == -1) count = this->indiceCount;
  
  
  // Re-Bind the buffers
  glBindBuffer(GL_ARRAY_BUFFER, this->vertexBuffer);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, this->indexBuffer);

  // Re-Calculate the attrib pointers.
  size_t offset = 0;
  glVertexAttribPointer(
    0, sizeof(glm::vec3) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  glEnableVertexAttribArray(0);

  offset += sizeof(glm::vec3) * this->verticeCount;
  glVertexAttribPointer(
    1, sizeof(glm::vec2) / sizeof(float_t),
    GL_FLOAT, GL_FALSE,
    0, (void *)offset
  );
  glEnableVertexAttribArray(1);

  // Render the elements.
  glDrawElements(
    drawMode, count, GL_UNSIGNED_INT, (void *)(sizeof(meshindice_t) * start)
  );
}

Mesh::~Mesh() {
  this->disposeBuffers();
}